"""
	Implements methods that are not part of the scripts themselves, but offer a few utility methods for things such as converting the output
	of breseq into a trajectory table or benchmarking the distance calculators.
"""